# Gitlab runner in Microsoft Azure

Gitlab runner is responsible for picking up tasks from gitlab, and running them. It can run on plethora of platforms, but the behaviour is always the same. By reading pipeline file `.gitlab-ci.yaml`, it creates a new container defined in "image" for every step, and runs the appropriate "script" phase. Script language depends on used image, but usually it's "sh", or similar. In language specific containers, it's usually the language in question, i.e. in `image: python:3` it would be the latest python v3. More details about runner is available at gitlab <https://docs.gitlab.com/runner/>.

There is more straight forward ways of building pipeline, but it's has been specially requested by private sector partners that we do this "manually". In this document, we detail phases for setting up gitlab runner either in ...

- In Azure virtual machine (recommended)
- In Azure Kubernetes cluster
- In local Linux machine (by adapting from "Install docker and gitlab runner").

Only one of these is required. Multiple runners in a hybrid environment is also possible.

[[_TOC_]]

## Get gitlab registration token

The first step is to check your project runner token. This is used to identify which project runner should pick tasks for. To find your project token go into your project repository in [gitlab](https://gitlab.jyu.fi).

1. From the left side menu, select **Settings** → **CI/CD**
2. In the **Runners** sections, select **[ Expand ]**.
    ![Show token](gitlab-runner-token-shown.png)
3. Make note of the runner **registration url** and **registration token**.

## Creating a runner in Azure

[Azure](https://portal.azure.com) is Microsoft cloud service. If you like interfaces designed by Nokia engineers in the 90's, you'll love azure. Reason we are using Azure is money 🦀. The teaching staff is not getting any money from Microsoft, but students are getting sponsoring money from Microsoft. To get 100$ worth of credits, use university shorthand email (i.e. username@jyu.fi) to sign in at <https://azure.microsoft.com/en-us/free/students/>

Google App Engine should also work, but it might require one team member to register with a credit card, even for trial (90 days) period. See resources document in Moodle.

## Option 1: Create Azure virtual machine

Azure virtual machine is not the most scalable solution, but it is noticeably cheaper than Kubernetes, as we can run it in much more limited environment. But as it is full blown operating system running on cloud it needs a bit manual tweaking with old school tools. And in real world, it would also need somebody to manage and update it.

### Create virtual machine

1. In [Azure portal frontpage](https://portal.azure.com) select from Azure Services "**➕ Create a resource**"

2. In "Create a resource" select from "Categories" → "**Compute**" → "**Virtual Machine**".

3. Set up linux virtual machine. "Ubuntu Server 20.04 LTS" is default and fine, but it's up to teams to select their preferred one. Just make sure docker is supported. Adventurous teams can see <https://docs.gitlab.com/runner/install/>, and even use windows - at their own peril.

    * Subscription:         `Azure for Students`
    * Resource group:       `TJTS5901-k22` (see previous section about creating resource group)
    * Virtual machine name: `gitlab-ci-runner` (or anything, distinct for azure identification)
    * Region:               `(Europe) Sweden Central` (see previous section about creating resource group)
    * Image:                `Ubuntu Server 20.04 LTS - Gen2`
    * Azure Spot Instance:  _Maybe_. As it uses unused capacity at discout prices, you could in theory run more powerful instance, but there is no quarantee of availability.
    * Size:                 `B1s` or `B1ms`. - In our case we are not building too complicated stuff, so cheap is fine. Course lasts about 3 months, so select size and price appropriately. Select from "**see all sizes**".

    * Authentication type:  `SSH public key`
    * Username:             `tjts5901-admin` or anything team prefers. For security reasons just don't use the default here, but make hacker to guess what you have used.
    * SSH public key source: `Generate new key pair`. Alternatively, if you have created ssh key in previously - like while setting up local environment -, you can upload and use it here.
    * Public inbound ports: `Allow selected ports`
    * Select inbound ports: `SSH (22)`

    * *(Optional)* Select **[ Next : Disks > ]** and change "Os disk type" to `Standard SSD` – or dependand of your workload.

    * Select **[ Review + create ]** → **[ Create ]**
    ![Azure review virtual machine creation](azure-review-vm.png)

    * Azure should notify you now about "Generate new key pair", and warn about loosing them if you don't download them. Save offered `.pem` file.

4. Wait for deployment to complete, and then select **[ Go to resource ]**
5. Congratulations, you are now a proud owner of your own Linux server 🐧 - or windows, if you were so inclined. 

### Connect to your virtual machine

In previous step we downloaded ssh key which allows us to enter the virtual machine.

1. In Azure portal, select your newly created virtual machine.

2. In "Virtual machine" view, in left menu select "**Connect**" from Settings group. 

3. Follow the guide provided in the SSH section. Note that the default key name, if you followed our example, is _not_ `tjts5901-admin.pem`, but `gitlab-ci-runner_key.pem`. If you downloaded the generated key, it should reside in Downloads folder now. You can define file the location for convenience purposes in the azure "Private key path" field and copy example command from Azure portal into your command prompt (requires windows 10, mac or linux):
    ![Azure SSH Command](azure-ssh-vm.png)
    ```
    C:\Users\User>ssh -i Downloads/gitlab-ci-runner_key.pem tjts5901-admin@51.107.183.129
    The authenticity of host '51.107.183.129 (51.107.183.129)' can't be established.
    ECDSA key fingerprint is SHA256:BqLlo12e29hE52dAapTUNzXF7gOek2gqewRJHk2w0x0.
    Are you sure you want to continue connecting (yes/no)? yes
    Warning: Permanently added '51.107.183.129' (ECDSA) to the list of known hosts.
    Welcome to Ubuntu 20.04.3 LTS (GNU/Linux 5.11.0-1023-azure x86_64)

    [-- blah blah blah --]

    tjts5901-admin@gitlab-ci-runner:~$
    ```

    Alternatively, you can use the cloud shell.
    * In the top toolbar, look for cloud shell icon next to search bar (`[>_]`). If it requests 
    * If storage creation is requested, select **[ Create storage ]**.
    * Upload your `.pem` file by looking "Upload/Download files" in the cloud shell toolbar.
    * run ssh command in the shell, replacing example ip with the one provided by "Connect" screen:
          ```
          $ chmod 400 gitlab-ci-runner_key.pem
          $ ssh -i gitlab-ci-runner_key.pem tjts5901-admin@51.107.183.129
          ```

4. Congratulations, you are now in your linux virtual machine.

### Install docker and gitlab runner

We want to use gitlab runner with docker executor. If you are using officially supported linux version, you can find gitlab runner installation instructions here <https://docs.gitlab.com/runner/install/linux-repository.html> and docker installation instructions <https://docs.docker.com/engine/install/ubuntu/>. 

But for shit and giggels, we use the docker method of running gitlab runner. In this way we have common environment for gitlab runners <https://docs.gitlab.com/runner/install/docker.html>. So we run docker, which runs gitlab runner, which builds docker containers, which we use to build our application containers. It's like onions; Layers all around and if you look it too closely, you'll start crying.

1. Connect into machine planned for building and testing (see "Connect to your virtual machine").

2. You can follow the manual steps detailed on the docker documentation. Or use the convenience script. Run in virtual machine:
    ```
    $ curl -fsSL https://get.docker.com -o get-docker.sh
    $ sudo sh get-docker.sh
    ```

3. Define local volume mount point for gitlab runner configs. Runner config files can be found on local system at `/srv/gitlab-runner/config`.
    ```
    $ sudo docker run -d --name gitlab-runner --restart always \
        -v /srv/gitlab-runner/config:/etc/gitlab-runner \
        -v /var/run/docker.sock:/var/run/docker.sock \
        gitlab/gitlab-runner:latest
    ```
    We can now check that the gitlab runner is running:
    ```
    $ sudo docker ps
    CONTAINER ID   IMAGE                         COMMAND                  CREATED          STATUS          PORTS     NAMES
    d7928b1658f1   gitlab/gitlab-runner:latest   "/usr/bin/dumb-init …"   39 seconds ago   Up 37 seconds             gitlab-runner

    ```

4. Register the runner with university GitLab instance to pick up jobs. More detailed documentation is available at <https://docs.gitlab.com/runner/register/index.html#docker>.
    * Kill (`-rm`) previos instance of gitlab runner (`gitlab/gitlab-runner`), and run (`run`) registering (`register`) with interactive shell (`-it`) - keeping the config volume mountpoint (`-v /srv/gitlab-runner/config:/etc/gitlab-runner`)
        ```
        $ sudo docker run --rm -it \
            -v /srv/gitlab-runner/config:/etc/gitlab-runner \
            gitlab/gitlab-runner register \
            --url "https://gitlab.jyu.fi" \
            --executor docker \
            --docker-image "docker:stable" \
            --run-untagged=true \
            --docker-volumes /var/run/docker.sock:/var/run/docker.sock
        ```
    * Fill out the details:
        * GitLab instance URL:  `https://gitlab.jyu.fi`
        * registration token:   see section "Get gitlab registration token"
        * description:          "gitlab runner running on Azure blah blah blah"
        * Tags:                 `x86_64,linux,docker,azure,B1s,vm,whatnot`
        * executor:             `docker`
        * default Docker image: `docker:stable`

5. Restart the gitlab runner with new config:
    ```
    $ sudo docker restart gitlab-runner
    ```
    `sudo` in front of everything means "**s**uper **u**ser **do**".

6. Check that the gitlab runner is registered for your project. In gitlab project look that **Settings** → **CI/CD** → **Runners** has your new runner in **Available specific runners**.
    ![Specific runner in gitlab](specific-runner-in-gitlab.png)

7. For now, allow runner to do all the available work. Next to the runner name select edit (the pen icon), and check that "**Run untagged jobs**" is true. Save changes with **[ Save changes ]**.
    ![Gitlab runner settings page](gitlab-runner-settings.png)

## Option 2: Set up a Kubernetes cluster

Kubernetes cluster is very much enterprise. It takes care of allocating and orchestrating machine setup. Bad news is that it's a very big and complicated beast. Kubernetes documentation states as requirements for master node - at a minimum - 4 cores and 16gb of memory. You'll be burning a lot of coal to just to keep the lights on. But as developer time is more expensive than computing time, automatic scaling, zone redundancy, provider agnostic nature and such make it a worthwhile deal. But those 100$ azure credits won't last a month.

1. In [Azure portal frontpage](https://portal.azure.com) select from Azure Services "**➕ Create a resource**"

2. In "Create a resource" select from "Categories" → "**Compute**" → "**Kubernetes Service**".

3. Set up the basics:
    * Subscription:         `Azure for Students`
    * Resource group:       `TJTS5901-k22` (see section "Create a resource group")

    * Cluster preset configuration: How rich are your parents? You can select using the preset configuration, or define machine in the node section later on. Standard machine - `DS2 v2` - costs about €83.72 month, and it's recommended to run 3 nodes for resiliency. The "cheaper" preselect machine B4ms costs €106.50, but should be enought to run one node, but without resiliancy guarantees or distributed availability.
    * cluster name:         `TJTS5901-k22` or similar distinct name.
    * Region:               `(Europe) Sweden central`
    * Availability zones:   `None` or if you're running critical application - which this is not - then within the limits of the budget. If you're using single pod, then availability zones are a moot point.
    * Kubernetes version:   `1.21.7 (default)`

    * Node side:            `Standard B4ms`
    * Scale method:         `Manual`
    * Node count:           `1`
    
    We are using only one node, as student subscription allows only limited amount of actual computers to be used. And we might need additional computing for different services. Or again, more adventurous and sysadmin oriented teams can use Kubernetes cluster to host __all__ the services they need. Or, the team can share the load and take responsibility for setting up different services on different subscriptions. Only money is the limit!

4. Select **[ Review + create ]**, and check that the "Role-based access control (RBAC)" is enabled. Then select **[ Create ]**
   ![Kubernetes settings in review](azure-kubernetes-review.png)

5. Go have a cup of coffee, it takes a couple of minutes. After deployment is done **[ Go to resource ]**.
   [You are now burning money](azure-kubernetes-overview.png)

### Gitlab runner in Kubernetes

To install gitlab runner in Kubernetes, we'll use helm chart. gitlab documentations is available at <https://docs.gitlab.com/runner/install/kubernetes.html>. In gitlab runner we then define docker-in-docker service, and in docker-in-docker we run docker images, which are responsible for building our application containers - our docker images. More adventurous types from affluent families can build containerd images, and deploy them in to their kubernetes cluester, but it's out of scope of this course. By defaulting to old kubernetes version, it's also possible to deploy docker images, but it will be soon deprecated.

1. Open cloud shell (or install helm locally).
    * In Azure portal, next to the search bar, is a cloud shell icon (`[>_]`). Cloud shell is terminal client, but on a web browser. If this is the first time you activate cloud shell, it requests that you create a storage account. At the current time it's created on different region where all our other stuff is, and on different resource group. But it's Ok, click on **[ Create storage ]**.

2. Register gitlab helm chart repository running on shell:
    ```
    $ helm repo add gitlab https://charts.gitlab.io
    ```

3. Gitlab has provided default configuration file at <https://gitlab.com/gitlab-org/charts/gitlab-runner/blob/main/values.yaml>. Documentation is available in <https://docs.gitlab.com/runner/install/kubernetes.html#configuring-gitlab-runner-using-the-helm-chart>. To make life a bit less painful, we can thank Mikael and Joonas from a previous course iteration for making necessary changes for us. Pre-defined helm chart is available at [`docs/tjts5901/azure-runner/values.yaml`](values.yaml). But some changes are necessary:
    * Set correct runner token in `runnerRegistrationToken`. see "Get gitlab registration token" for details.
    * Advanced bonus: set up azure caching.

4. Upload your modified `values.yaml` into the cloud shell storage.
   * In the cloud shell dialog, look for "Upload/Download files" button on toolbar. Select your modified `values.yaml` file.


5. Authenticate with Kubernetes cluster:
   ```
   $ az aks get-credentials --resource-group TJTS5901-k22 --name TJTS5901-k22
   ```

6. Deploy helm chart into Kubernetes
   ```
   helm install --namespace default gitlab-runner -f values.yaml gitlab/gitlab-runner
   ```
   If you update `values.yaml`, you can re-deploy it with
   ```
   helm upgrade gitlab-runner --namespace default -f values.yaml gitlab/gitlab-runner
   ```
7. Check that the gitlab runner is registered for your project. In gitlab project look that **Settings** → **CI/CD** → **Runners** has your new runner in **Available specific runners**.
    ![Specific runner in gitlab](kubernetes-specific-runner-in-gitlab.png)

8. For now, allow runner to do all the available work. Next to the runner name select edit (the pen icon), and check that "**Run untagged jobs**" is true. While there, you might want to add some tags as "kubernetes, azure, x86_64, whatnot" and change the description. Save changes with **[ Save changes ]**.
    ![Gitlab runner settings page](gitlab-runner-settings.png)

### Set pipeline to use Docker-in-Docker

To build docker images, we'll use docker-in-docker <https://docs.gitlab.com/ee/ci/docker/using_docker_build.html>. In a single machine instance, we could use the host provided docker with a bit volume mounting magic, but in a distributed environment, we need to set up node/service to build things.

1. In project `.gitlab-ci.yml` add into `services` section `docker:stable-dind`

2. Add necessary environment variables for remote docker connection. In `variables` section add: `DOCKER_HOST: tcp://docker:2376`, `  DOCKER_TLS_CERTDIR: "/certs"`, `DOCKER_TLS_VERIFY: 1` and `DOCKER_CERT_PATH: "$DOCKER_TLS_CERTDIR/client"`

For your convenience minimal example is provided in (`docs/tjts5901/azure-runner/gitlab-ci-dind.yml`)[gitlab-ci-dind.yml]
