# Deploying Container to Cloud

For running container in cloud Azure offers a "Web App" service, and Google has "App Engine". Both provide ready-made environments for different languages or can be used to run custom images. Documentation for setting up Azure App is at <https://docs.microsoft.com/en-us/azure/app-service/tutorial-custom-container>  and <https://cloud.google.com/appengine/docs/flexible/custom-runtimes/quickstart>

## Deploying container to Azure

1. Create new web app resource.
    * In the [Azure portal](https://portal.azure.com), create a new reosource by selecting "**➕ Create a resource**"
    * From "**Web**" category select "**Web App**"

2. For Basics:
    * Subscription:         `Azure for Students`
    * Resource group:       `TJTS5901-k22`

    * Name:                 (example) `teamname-dogmap` or other suitable name.
    * Publish:              `Docker Container`
    * Operating System:     `Linux`
    * Region:               `North Europe` - North europe offers Free (as in free beer) instance `F1`

    * Linux Plan:           (example) `TJTS5901-WebApp`
    * SKU and size:         `Free F1` - notice the limitations.

3. Select the **[ Next : Docker > ]**, and in "Docker" page select:
    * Options:              `Single Container`
    * Image source:         `Azure Container Registry`

    * Registry:             `your-previously-created-registry` (see [Container Registry](../docker-image/Container%20registry.md))
    * Image:                `<your-dogmap>` If no image is available, you haven't pushed anything yet. See [Container Registry](../docker-image/Container%20registry.md) information about pushing an image to a registry.
    * Tag:                  `<your-tag>`. See above.

4. Select **[ Review + create ]** → **[ Create ]** 
    ![Webapp setup review](azure-webapp-review.png)

5. After deployment is completed **[ Go to resource ]**

6. Define webapp port by using the left side menu, **Settings** → **Configuration** → **+  New application setting**.
    * Name:                 `WEBSITES_PORT`
    * Value:                `5000`

    Flask runs by default on 5000, and Azure expects the port to be standard http port 80. (For historical reasons, access to ports under 1024 were limited).

7. Select **[ OK ]** and from the top of the modal select "**Save**"
    ![Azure webapp setting](azure-webapp-settings.png)

(Optional): Create managed identity to hide docker login information. See <https://docs.microsoft.com/en-us/azure/app-service/tutorial-custom-container?pivots=container-linux#configure-app-service-to-deploy-the-image-from-the-registry>

### Itegrating automatic deployment

Using Azure provided webhook, we can create a "trigger" that every time a new image is added to the registry, a new container is deployed.

1. In the web app service page, select from the left side menu "**Deployment Center**".

2. In the Deployment Center Settings page, change "**Continuous deployment**" to **On**. Remember to click **[ Save ]** 

3. From the same page copy "Webhook URL"
    ![Continuous deployment toggle switch](azure-webapp-cd.png)

4. Find your teams "Container registry" instance from Azure portal.

5. Select "**Webhooks**" from the left side menu, and click "**➕ Add**"

6. Fill out "Create webhook" page.
    * Webhook name:             (example) `webappDeployment`
    * Location:                 `Sweden Central`
    * Service URI:              Paste "Webhook URL" copied from web app service "Deployment Center" page.
    * Actions:                  `push`
    * Scope:                    (example) `dogmap:latest` - logical to use the same as used in "Deployment Center" page. Only images tagged as such are deployed.

    Select **[ Create ]**
    ![Hook settings](azure-acr-hook.png)

7. (Optional) Test.
    1. See build date of your web app by visiting `/build` i.e. `https://<your-webapp>.azurewebsites.net/build`

    2. Build and tag a new image:
        ```
        docker build --pull --rm -f "Dockerfile" -t dogmap .
        docker tag dogmap <your-acr-id>.azurecr.io/dogmap:latest
        ```

    3. Push new image
        ```
        docker push tjts5901.azurecr.io/dogmap
        ```

    4. Wait for it (minute or two)

    5. Visit same `/build` address to see if date has changed. It should have.
