# Docker Containers

## Container registry / hub

A container registry is a place where we can publish our own images. Teams can use either Docker provided registry, or Azure Container Registry to keep their containers private.

## Create Azure container registry

1. In the [azure portal](https://portal.azure.com), select "**[ ➕ Create a resource ]**
   
2. From categories, select "**Containers**" → "**Container Registry**

3. Fill out project details:
   * Subscription:      `Azure for Students`
   * Resource group:    `TJTS5901-k22` (see ["Create a resource group"](azure-runner/Gitlab%20runner%20in%20Microsoft%20Azure.md))
   * Registry name:     `your-team-name`
   * Location:          `Sweden Central`
   * SKU:               `Basic`

4. Select **[ Review + create ]** → **[ Create ]**

5. After deployment is completed, select **[ Go to resource ]**

6. In the left side menu, select "Access Keys" and enable "**Admin user**".

7. Make note of the Access key page information, and use it to login into registry. Using docker command line either in cloud shell or in local:
   ```
   $ docker login <Login server> -u "<Registry name>"
   ```
   ![acr docker login example](acr-login.png)


## Building, running, pushing and pulling

For example, application is named `dogmap`, and registry is at `tjts5901.azurecr.io`

* Building docker image from Dockerfile:
   ```
   $ docker build --pull --rm -f "Dockerfile" -t dogmap "."
   ```

* Run the container in foreground (`-it`).
   ```
   $ docker run --rm -it dogmap
   ```

* Login into container registry (see "Access Keys" page from container registry).
   ```
   $ docker login tjts5901.azurecr.io -u "tjts5901" -p "<password>"
   ```

* Tag built image to indicate it's "same" as the one in container registry:
   ```
   $ docker tag dogmap tjts5901.azurecr.io/dogmap:latest
   ```
   <https://docs.docker.com/engine/reference/commandline/tag/>

* Push new image to registry:
   ```
   $ docker push tjts5901.azurecr.io/dogmap:latest
   ```

* pull latest image from registry:
   ```
   $ docker pull tjts5901.azurecr.io/dogmap:latest
   ```
