from flask import Flask
from flask.templating import render_template


def create_app() -> Flask:
    """
    Create new flask application.
    """

    app = Flask(__name__)

    return app


app = create_app()


@app.route("/")
def index():
    """
    Default route for intex page.
    """

    # Define page attributes for template
    page = {
        'title': "Minnan testisivu."
    }

    # coordinates to Agora
    agora_location = {
        'lat': 60.276367,
        'lon': 24.750278
    }

    # List for markers
    markers = []

    # Add marker for Agora
    markers.append({
        'loc': agora_location,
        'description': "Home Sweet Home"
    })

    # Render template takes template name first, and variables in template as
    # additional params.
    return render_template("map.j2.html",
                           page=page,
                           map_center=agora_location,
                           markers=markers
                           )


@app.route("/build")
def image_build():
    """
    Display build date.

    BUILD.txt file is created at docker image creation stage.
    """
    try:
        with open("BUILD.txt") as fd:
            build = fd.read()
    except FileNotFoundError:
        build = "Build info not found"

    return build
